from networkx.generators.random_graphs import gnm_random_graph


def gen_graph(n_problems, size, seed=None):
    if seed is not None:
        return [
            gnm_random_graph(*size, seed=seed)
            for _ in range(n_problems)
        ]
    return [gnm_random_graph(*size) for _ in range(n_problems)]



import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


np.seterr(divide='ignore', invalid='ignore')

mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')

# Source: https://stackoverflow.com/questions/30079590/
# For continuous:
# n = 10
# color = plt.cm.viridis(np.linspace(0, 1, n))
# mpl.rcParams['axes.prop_cycle'] = cycler.cycler('color', color)
# For discrete:
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)


def plot_loss(kv):
    fig, ax = plt.subplots(figsize=(4, 3))

    tags = {
        'a19_2_r2': 'Adjacency matrix',
        'a19_2_r2_gen_edges2': 'Edge list'
    }

    def sub_plot(ax, key, arr):
        arr = arr[:, :500]

        next(ax._get_lines.prop_cycler)['color']

        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    for i, (k, v) in enumerate(kv.items()):
        # sub_plot(k, v[0])  # This is eval
        sub_plot(ax, k, v[1])  # This is train
        ax.legend(frameon=False)
        if i == 0:
            ax.set_ylabel("Train Loss")
        ax.set_xlabel("Epoch")
        ax.set_ylim([-.5, 10])
        # ax.set_title("Maximum Cut")

    plt.tight_layout()
    plt.show()
    fig.savefig('mc_comp_loss.png')
    fig.savefig('mc_comp_loss.pdf')


def plot_r2(kv):
    fig, ax = plt.subplots(figsize=(4, 3))

    tags = {
        'a19_2_r2': 'Adjacency matrix',
        'a19_2_r2_gen_edges2': 'Edge list'
    }

    def sub_plot_r2(i, ax, key, arr):
        arr = arr[:, :500]

        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

        idx = np.argmax(mean)

        if i == 1:
            ax.add_artist(
                mpl.patches.Ellipse(
                    (idx, mean[idx]),
                    width=15.,
                    height=.05,
                    fc='none',
                    ec='r'
                )
            )

        print("Maximum R2", key, mean[idx], "+-", mean[idx] - ci[0][idx])

    for i, (k, v) in enumerate(kv.items()):
        # sub_plot(k, v[0])  # This is eval
        sub_plot_r2(i, ax, k, v[2])  # This is R**2
        ax.legend(frameon=False)
        if i == 0:
            ax.set_ylabel(r'$R^2$')
        ax.set_xlabel("Epoch")
        ax.set_ylim([0, 1.1])
        # ax.set_title(tags[k])

    plt.tight_layout()
    plt.show()
    fig.savefig('mc_comp_edges_r2.png')
    fig.savefig('mc_comp_edges_r2.pdf')


def run():
    with open('mc_comp_edges.pickle', 'rb') as f:
        kv = pickle.load(f)
    plot_loss(kv)
    plot_r2(kv)


if __name__ == '__main__':
    run()



import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


np.seterr(divide='ignore', invalid='ignore')

mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')

# Source: https://stackoverflow.com/questions/30079590/
# For continuous:
# n = 10
# color = plt.cm.viridis(np.linspace(0, 1, n))
# mpl.rcParams['axes.prop_cycle'] = cycler.cycler('color', color)
# For discrete:
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)


def plot(kv):
    fig, axs = plt.subplots(1, 5, figsize=(10, 2.0))
    axs = axs.flatten()

    tags = {
        'a19_2_r2': 'Maximum Cut',
        'np19_LONG_r2': 'Number Partitioning',
        'gc1_r2': 'Graph Coloring',
        'mvc3_r2': 'Minimum Vertex Cover',
        'tsp2_r2': ' Traveling Salesman'
        # 'a19_gen_edges': 'Maximum Cut - edges'
    }

    def sub_plot(ax, key, arr):
        arr = arr[:, :1000]
        next(ax._get_lines.prop_cycler)['color']

        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    for i, (k, v) in enumerate(kv.items()):
        # sub_plot(k, v[0])  # This is eval
        sub_plot(axs[i], k, v[1])  # This is train
        # ax.legend()
        if i == 0:
            axs[i].set_ylabel("Train Loss")
        axs[i].set_xlabel("Epoch")
        axs[i].set_ylim([-.5, 10])
        axs[i].set_title(tags[k])

    plt.tight_layout()
    plt.show()
    fig.savefig('reverse_loss.png')
    fig.savefig('reverse_loss.pdf')


def plot_r2(kv):
    fig, axs = plt.subplots(1, 5, figsize=(10, 2.0))
    axs = axs.flatten()

    tags = {
        'a19_2_r2': 'Maximum Cut',
        'np19_LONG_r2': 'Number Partitioning',
        'gc1_r2': 'Graph Coloring',
        'mvc3_r2': 'Minimum Vertex Cover',
        'tsp2_r2': ' Traveling Salesman'
        # 'a19_gen_edges': 'Maximum Cut - edges'
    }

    def sub_plot_r2(ax, key, arr):
        # arr = arr[:, :5000]

        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)
        print("R2", key, mean[-1], "+-", mean[-1] - ci[0][-1])

    for i, (k, v) in enumerate(kv.items()):
        # sub_plot(k, v[0])  # This is eval
        sub_plot_r2(axs[i], k, v[2])  # This is R**2
        # axs[i].legend()
        if i == 0:
            axs[i].set_ylabel(r'$R^2$')
        axs[i].set_xlabel("Epoch")
        axs[i].set_ylim([0, 1.1])
        axs[i].set_title(tags[k])

    plt.tight_layout()
    plt.show()
    fig.savefig('reverse_r2.png')
    fig.savefig('reverse_r2.pdf')


def run():
    with open('reverse_losses.pickle', 'rb') as f:
        kv = pickle.load(f)
    plot(kv)
    plot_r2(kv)


if __name__ == '__main__':
    run()



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, len_, req_len=None):
    r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=req_len)
    eval_loss = aggregate_single(paths, 'Loss/Eval', min_len, req_len=req_len)
    return r2, eval_loss


def run():
    min_len = 200
    req_len = 10
    base_paths = ['../runs/']
    problems = [
        ('qk7_6x6_high30_1M', 10),
        ('qk7_6x6_high10_1M', 10),
        ('qk7_6x6_high30_1M_multiply', 10),
        ('qk8_16x16_high30_1M', 10),
        ('qk8_16x16_high30_1M_multiply', 10),
        ('qk8_6x6_high10_1M', 10)
    ]
    kv = {}
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        data = aggregate(paths, min_len, problem[1], req_len=req_len)
        kv[problem[0]] = data

    problems = [
        ('qk10', 10),
        ('qk10_norm', 10),
        ('qk10_high10', 10),
        ('qk10_high10_1', 10),
        ('qk10_norm_random_P', 10)
    ]
    req_len = 200
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        data = aggregate(paths, min_len, problem[1], req_len=req_len)
        kv[problem[0]] = data

    with open('qk.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import random
import collections
import numpy as np
from qubo_nn.problems.problem import Problem


class Max2SAT(Problem):
    def __init__(self, cfg, clauses, n_vars):
        self.clauses = clauses
        self.n_vars = n_vars

    def gen_qubo_matrix(self):
        n = self.n_vars
        Q = np.zeros((n, n))

        # Clauses is a list of 2-tuples of 2-tuples.
        # Eg: [((1, True), (2, True)), ((1, True), (2, False))]
        # So we loop through all tuples and find the type of it.
        # There are 4 types: (True, True), (True, False), etc.
        for clause in self.clauses:
            idx1 = clause[0][0]
            idx2 = clause[1][0]
            val1 = clause[0][1]
            val2 = clause[1][1]

            if val1 and val2:
                Q[idx1][idx1] -= 1
                Q[idx2][idx2] -= 1
                Q[idx1][idx2] += .5
                Q[idx2][idx1] += .5
            elif val1 and not val2:
                Q[idx2][idx2] += 1
                Q[idx1][idx2] -= .5
                Q[idx2][idx1] -= .5
            elif not val1 and val2:
                Q[idx1][idx1] += 1
                Q[idx1][idx2] -= .5
                Q[idx2][idx1] -= .5
            elif not val1 and not val2:
                Q[idx1][idx2] += .5
                Q[idx2][idx1] += .5
        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size=(20, 25), **kwargs):
        check_duplicates = cfg["problems"]["M2SAT"].get("check_duplicates", False)
        if check_duplicates:
            print("M2SAT gen_problems: Checking for duplicates.")

        # size: First is n_vars, second is number of clauses.
        # Want something like this:
        # [((1, True), (2, True)), ((1, True), (2, False))]
        n_vars = size[0] - 1
        problems = []
        for _ in range(n_problems):
            problem = []

            assoc = collections.defaultdict(list)

            # Eg: 25 clauses
            for _ in range(size[1]):
                # Each has TWO tuples!! But random vars and random value
                # (True/False).
                idx1 = int(round(random.random() * n_vars))
                idx2 = int(round(random.random() * n_vars))
                val1 = bool(random.getrandbits(1))
                val2 = bool(random.getrandbits(1))

                # TODO: With the checks below we sometimes end up with less than the desired # of clauses.

                if check_duplicates:
                    # TODO Added Mar 19 06:45
                    if idx1 == idx2:
                        continue

                    if not val1:
                        idx1 = -idx1
                    if not val2:
                        idx2 = -idx2

                    cont = False
                    for x in assoc[-idx1]:
                        for y in assoc[-idx2]:
                            if -x == y:
                                cont = True

                    if cont:
                        continue

                    assoc[idx1].append(idx2)
                    assoc[idx2].append(idx1)

                clause = (
                    (abs(idx1), val1),
                    (abs(idx2), val2)
                )
                clause = sorted(clause)

                if check_duplicates:
                    if clause in problem:
                        continue

                    # TODO Added Mar 19 07:45
                    other_clause1 = [clause[0], (clause[1][0], not clause[1][1])]
                    other_clause2 = [(clause[0][0], not clause[0][1]), clause[1]]
                    other_clause3 = [(clause[0][0], not clause[0][1]), (clause[1][0], not clause[1][1])]
                    if other_clause1 in problem or other_clause2 in problem or other_clause3 in problem:
                        continue

                problem.append(clause)
            problems.append(sorted(problem))
        return [{"clauses": problem, "n_vars": size[0]} for problem in problems]



import itertools
import numpy as np
from qubo_nn.problems.problem import Problem
from qubo_nn.problems.util import gen_graph


class MaxClique(Problem):
    def __init__(self, cfg, graph):
        self.graph = graph

    def gen_qubo_matrix(self):
        n = self.graph.order()

        Q = np.zeros((n, n))

        for i in range(n):
            for j in range(n):
                if i == j:
                    Q[i][i] = -1
                if i < j and (i, j) not in self.graph.edges:
                    Q[i][j] = 2
                    Q[j][i] = 2

        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size, seed=None, **kwargs):
        graphs = gen_graph(n_problems, size, seed)
        return [{"graph": graph} for graph in graphs]



import numpy as np
from qubo_nn.problems.problem import Problem
from qubo_nn.problems.util import gen_graph


class MaxCut(Problem):
    def __init__(self, cfg, graph):
        self.graph = graph

    def gen_qubo_matrix(self):
        n = self.graph.order()
        nodes = list(self.graph.nodes)

        # dtype 'b' would be nice here..
        Q = np.zeros((n, n), dtype=np.dtype(np.int32))
        for edge in self.graph.edges:
            idx1 = nodes.index(edge[0])
            idx2 = nodes.index(edge[1])
            Q[idx1][idx1] += 1
            Q[idx2][idx2] += 1
            Q[idx1][idx2] -= 1
            Q[idx2][idx1] -= 1
        return -Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size, seed=None, **kwargs):
        graphs = gen_graph(n_problems, size, seed)
        return [{"graph": graph} for graph in graphs]


class MaxCutMemoryEfficient(Problem):
    def __init__(self, cfg, edge_list, n):
        self.edge_list = edge_list
        self.n = n

    def gen_qubo_matrix(self):
        Q = np.zeros((self.n, self.n), dtype=np.dtype('b'))
        for edge in self.edge_list:
            idx1 = edge[0]
            idx2 = edge[1]
            Q[idx1][idx1] += 1
            Q[idx2][idx2] += 1
            Q[idx1][idx2] -= 1
            Q[idx2][idx1] -= 1
        return -Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size, seed=None, **kwargs):
        data = []
        for i in range(n_problems):
            if i % 100000 == 0:
                print(i)
            data.append(list(gen_graph(1, size, seed)[0].edges))
        return [{"edge_list": graph, "n": size[0]} for graph in data]



import pickle
import numpy as np
from MulticoreTSNE import MulticoreTSNE as TSNE
from qubo_nn.data import LMDBDataLoader
from qubo_nn.config import Config


cfg_id = '30_gen4'
cfg = Config('../').get_cfg(cfg_id)
cfg["use_big"] = False
lmdb_loader = LMDBDataLoader(cfg, reverse=False, base_path='../')

X = []
y = []
for i, data in enumerate(lmdb_loader.train_data_loader):
    if i > 87:  # 88 batches á 500 = 44k (from total of 560k), so ~8%
        break
    X.extend(data[0].tolist())
    y.extend(data[1].tolist())

X = np.array(X)
X = X.reshape(-1, 64**2)
print(X.shape)

for i in [10, 20, 30, 50, 70, 100, 200, 500, 1000]:
    tsne = TSNE(
        n_jobs=10,
        n_iter=5000,
        perplexity=i,
        # perplexity=500.,  # Best.
        verbose=1
    )
    Y = tsne.fit_transform(X)

    with open('tsne_30_gen4_data%d.pickle' % i, 'wb+') as f:
        pickle.dump((Y, y), f)



import pickle
import collections

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
# n = 6
# color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
# mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)


def gen_table(kv):
    def calc_ci(key, arr):
        arr = arr[~np.isnan(arr)]
        arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        range_ = round(mean - ci[0], 4)
        mean = round(mean, 4)
        return mean, range_

    # First, aggregate.
    kv_new = collections.defaultdict(list)
    for k, v in kv.items():
        kv_new[k[:-1]].extend(v)

    for k, v in kv_new.items():
        # arr = arr[:, 1:201]
        v = np.array(v)
        mean, range_ = calc_ci(k, v[1].max(axis=1))  # r2
        print(k, "R2", "%.3f" % mean, "+-", "%.3f" % range_)
        mean, range_ = calc_ci(k, v[2].min(axis=1))  # eval loss
        print(k, "Eval Loss", "%.5f" % mean, "+-", "%.5f" % range_)


def plot(kv):
    tags = {
        'qa_N_9_norm2_1': '9x9',
        'qa_N_16_norm': '16x16',
        'qa_N_64_norm': '64x64',
        'qa_N_100_norm': '100x100',
        'qa_N_144_norm': '144x144'
    }

    fig, axs = plt.subplots(1, 1, figsize=(5, 3.5))

    def calc_ci(ax, key, arr):
        # arr = arr[~np.isnan(arr)]
        # arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    # First, aggregate.
    kv_new = collections.defaultdict(list)
    for k, v in kv.items():
        kv_new[k[:-1]].extend(v)

    for k, v in kv_new.items():
        if k not in tags:
            continue
        v = np.array(v)
        calc_ci(axs, k, v[1][:, :100])  # r2

        axs.legend()
        axs.set_ylabel(r'$R^2$')
        axs.set_xlabel("Epoch")

        # # Shrink current axis by 20%
        # box = axs.get_position()
        # axs.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        # # Put a legend to the right of the current axis
        # axs.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.tight_layout()
    plt.show()
    fig.savefig('qa_comp.png')
    fig.savefig('qa_comp.pdf')


def plot_datasize(kv):
    tags = {
        'qa_N_16_norm': '16x16, 200k',
        'qa_N_16_norm2_1': '16x16, 1M',
        'qa_N_16_norm2_4': '16x16, 4M'
    }

    fig, axs = plt.subplots(1, 1, figsize=(5, 3.5))

    def calc_ci(ax, key, arr):
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    # First, aggregate.
    kv_new = collections.defaultdict(list)
    for k, v in kv.items():
        kv_new[k[:-1]].extend(v)

    for k, v in kv_new.items():
        if k not in tags:
            continue
        v = np.array(v)
        calc_ci(axs, k, v[1][:, :50])  # r2

        axs.legend()
        axs.set_ylabel(r'$R^2$')
        axs.set_xlabel("Epoch")

    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.tight_layout()
    plt.show()
    fig.savefig('qa_comp_datasize.png')
    fig.savefig('qa_comp_datasize.pdf')


def plot_special_loss(kv):
    tags = {
        'qa_special_loss1': 'QA 16x16'
    }

    fig, axs = plt.subplots(1, 1, figsize=(5, 3.5))

    def calc_ci(ax, key, arr):
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    for k in tags.keys():
        v = np.array(kv[k])
        calc_ci(axs, k, v[2][:, :50])  # eval loss

        axs.legend()
        axs.set_ylabel('Loss')
        axs.set_xlabel("Epoch")

    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.tight_layout()
    plt.show()
    fig.savefig('qa_comp_special_loss.png')
    fig.savefig('qa_comp_special_loss.pdf')


def run():
    with open('qa_comp.pickle', 'rb') as f:
        kv = pickle.load(f)
    gen_table(kv)
    plot(kv)
    plot_datasize(kv)
    plot_special_loss(kv)


if __name__ == '__main__':
    run()



from qubo_nn.nn.models import Optimizer
from qubo_nn.nn.models import ReverseOptimizer
from qubo_nn.nn.models import AutoEncoderOptimizer
from qubo_nn.nn.models import RNNOptimizer
from qubo_nn.nn.models import A3Optimizer
from qubo_nn.nn.models import Resistance1
from qubo_nn.nn.models import Resistance2
from qubo_nn.nn.models import QbsolvOptimizer
from qubo_nn.nn.models import RedAEOptimizer



# TODO: I probably won't use abc here due to a possibility of using Cython.

class Problem:
    def gen_qubo_matrix(self):
        pass

    # TODO Make sure overriding this works.
    @classmethod
    def gen_problems(self, cfg, n_problems, size=20, **kwargs):
        pass



import pickle
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib as mpl


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')


def truncate_colormap(cmapIn='jet', minval=0.0, maxval=1.0, n=100):
    '''truncate_colormap(cmapIn='jet', minval=0.0, maxval=1.0, n=100)'''
    cmapIn = plt.get_cmap(cmapIn)

    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmapIn.name, a=minval, b=maxval),
        cmapIn(np.linspace(minval, maxval, n)))

    return new_cmap


with open('qubo_map.pickle', 'rb') as f:
    medians, singles = pickle.load(f)

# Total length is 900k
problems = ["NP", "MC", "MVC", "SP", "M2SAT", "SPP", "GC", "QA", "QK"]

# Plot.
cmap_mod = truncate_colormap('Greens', minval=.3, maxval=.99)


def plot(data, fname):
    fig, axs = plt.subplots(3, 3, figsize=(9, 8), constrained_layout=True)
    for i in range(3):
        for j in range(3):
            idx = i * 3 + j
            print(idx)
            im = axs[i][j].imshow(data[idx], cmap=cmap_mod, vmin=-1, vmax=1)
            cbar = axs[i][j].figure.colorbar(im, ax=axs, aspect=60)
            axs[i][j].set_title(problems[idx])
            axs[i][j].set_xticks([])
            axs[i][j].set_yticks([])
    fig.savefig(fname + ".png", bbox_inches='tight')
    plt.show()


plot(medians, "qubo_map_medians")
plot(singles, "qubo_map_singles")



import numpy as np
from qubo_nn.problems.problem import Problem
from qubo_nn.problems.util import gen_graph


class MinimumVertexCover(Problem):
    def __init__(self, cfg, graph, P=8):
        self.graph = graph
        self.P = P

    def gen_qubo_matrix(self):
        n = self.graph.order()
        nodes = list(self.graph.nodes)

        Q = np.zeros((n, n))
        for i in range(n):
            Q[i][i] += 1

        for edge in self.graph.edges:
            idx1 = nodes.index(edge[0])
            idx2 = nodes.index(edge[1])
            Q[idx1][idx1] -= self.P
            Q[idx2][idx2] -= self.P
            Q[idx1][idx2] += self.P / 2.
            Q[idx2][idx1] += self.P / 2.
        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size=(20, 25), seed=None, **kwargs):
        graphs = gen_graph(n_problems, size, seed)
        return [{"graph": graph} for graph in graphs]



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, cutoff=False):
    arr_eval = aggregate_single(paths, 'Loss/Eval', min_len)
    arr_train = aggregate_single(paths, 'Loss/Train', min_len)
    if cutoff:
        arr_r2 = aggregate_single(paths, 'Custom/R2', min_len, cutoff=6)
    else:
        arr_r2 = aggregate_single(paths, 'Custom/R2', min_len)
    return arr_eval, arr_train, arr_r2


def run():
    min_len = 500
    base_path = '../runs3/'
    problems = ['m2sat_16x16_5_F_v2', 'm2sat_16x16_5_F_v2_no_dupl_chk']
    kv = {}
    for problem in problems:
        print(problem)
        paths = glob.glob('../runs3/*-' + problem)

        paths = [path for path in paths if path not in [
            # '../runs/21-04-03_13:01:10-8112457-datolith.cip.ifi.lmu.de-m2sat_16x16_5_F_v2',
            '../runs3/21-04-02_07:40:21-3067619-feueropal.cip.ifi.lmu.de-m2sat_16x16_5_F_v2'
        ]]

        print(paths)
        if problem == 'm2sat_16x16_5_F_v2_no_dupl_chk':
            data = aggregate(paths, min_len, True)
        else:
            data = aggregate(paths, min_len, False)
        kv[problem] = data

    with open('m2sat_three_case.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import os
import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


NAME = os.path.splitext(os.path.basename(__file__))[0][5:]

mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
# n = 6
# color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
# mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)

PLOT_TAGS = [
    {
        'a3_1': 'AutoEncoder + Similarity Loss'
    },
    {
        'a3_3': 'AutoEncoder + Similarity Loss + QBSolv Loss'
    },
    {
        'a3ng': 'AutoEncoder + Similarity Loss + QBSolv Loss [100k]'
    },
    {
        'a3ng_v2_e1': 'AutoEncoder + Similarity Loss + QBSolv Loss [1M]'
    }
]
PLOT_NAMES = [
    'a3_1',
    'a3_3',
    'a3ng',
    'a3ng_v2_e1'
]
PLOT_LIMS = [
    (-0.05, 1.01, 50000),
    (5, 16, 50000),
    (5, 16, 50000),
    (5, 16, 50000)
]


def plot(kv, tags, name, lims):
    fig, axs = plt.subplots(1, 1, figsize=(5, 3.75))

    def calc_ci(ax, key, arr):
        # arr = arr[~np.isnan(arr)]
        # arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        print(mean[-1], "+-", mean[-1] - ci[0][-1])

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    for k in tags:
        print(k)
        v = kv[k]
        v = np.array(v)
        calc_ci(axs, k, v[1][:, :lims[-1]])  # train loss

        axs.legend()
        axs.set_ylabel('Train Loss')
        axs.set_xlabel("Epoch")

    # plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.legend(frameon=False)
    plt.ylim(lims[0:2])
    plt.tight_layout()
    plt.show()
    fig.savefig(NAME + '_' + name + '.png')
    fig.savefig(NAME + '_' + name + '.pdf')


def run():
    with open(NAME + '.pickle', 'rb') as f:
        kv = pickle.load(f)
    for plot_tags, plot_name, lims in zip(PLOT_TAGS, PLOT_NAMES, PLOT_LIMS):
        plot(kv, plot_tags, plot_name, lims)


if __name__ == '__main__':
    run()



import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt
import matplotlib.colors as colors


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')


def truncate_colormap(cmapIn='jet', minval=0.0, maxval=1.0, n=100):
    cmapIn = plt.get_cmap(cmapIn)

    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmapIn.name, a=minval, b=maxval),
        cmapIn(np.linspace(minval, maxval, n)))

    return new_cmap


cmap_mod = truncate_colormap('Greens', minval=.5, maxval=.99)


def plot_confusion(id_):
    with open('tot_mc_100_genX_%s.pickle' % id_, 'rb') as f:
        arr1, mc_tables = pickle.load(f)

    # mc_table = np.mean(mc_tables, axis=0)
    mean = np.mean(mc_tables, axis=0)
    mean = mean.reshape(14 * 14)
    var = st.sem(mc_tables, axis=0)
    var = var.reshape(14 * 14)

    ci = st.t.interval(
        0.95,
        len(mc_tables) - 1,
        loc=mean,
        scale=var
    )

    mean = mean.reshape((14, 14))
    ci = mean - ci[0].reshape((14, 14))

    fig, ax = plt.subplots(figsize=(7, 7))
    plt.ylabel('True label')
    plt.xlabel('Predicted label')

    for i in range(mean.shape[0]):
        for j in range(mean.shape[0]):
            if np.isnan(ci[i][j]):
                ci[i][j] = 0.
            ax.text(
                j,
                i,
                '%.02f \n± %.02f' % (mean[i][j], ci[i][j]),
                ha="center",
                va="center",
                color="w",
                fontsize=9
            )

    problems = ["NP", "MC", "MVC", "SP", "M2SAT", "SPP", "GC", "QA", "QK", "M3SAT", "TSP", "GI", "SGI", "MCQ"]  # noqa
    plt.xticks(list(range(len(problems))), problems)
    plt.yticks(list(range(len(problems))), problems)

    plt.tight_layout()
    ax.imshow(mean, cmap=cmap_mod, vmin=0, vmax=1)
    plt.savefig('confusion_100_genX_%s.png' % id_)
    plt.savefig('confusion_100_genX_%s.pdf' % id_)


def plot(id_):
    with open('tot_mc_100_genX_%s.pickle' % id_, 'rb') as f:
        arr1, confusion_matrix = pickle.load(f)

    fig, ax = plt.subplots(figsize=(3.5, 3))

    def sub_plot(arr, col):
        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])

        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, color=col)
        ax.fill_between(x, ci[0], ci[1], color=col, alpha=.1)

        print(mean[-1], "+-", mean[-1] - ci[0][-1])

    sub_plot(arr1, 'c')
    plt.ylabel("Misclassification rate")
    plt.xlabel("Epoch")
    plt.tight_layout()
    plt.show()
    fig.savefig('tot_mc_100_genX_%s.png' % id_)
    fig.savefig('tot_mc_100_genX_%s.pdf' % id_)


def run():
    plot_confusion('100_genX')
    plot('100_genX')
    plot('100_genX_2')


if __name__ == '__main__':
    run()



import unittest
import networkx
from qubo_nn.problems import MinimumVertexCover


class TestMinimumVertexCover(unittest.TestCase):
    def test_gen_qubo_matrix(self):
        """Test whether a correct QUBO is generated.

        Test case from: https://arxiv.org/pdf/1811.11538.pdf
        """
        graph = networkx.Graph(
            [(1, 2), (1, 3), (2, 4), (3, 4), (4, 5), (3, 5)]
        )
        problem = MinimumVertexCover({}, graph)
        matrix = problem.gen_qubo_matrix()
        want = [
            [-15, 4, 4, 0, 0],
            [4, -15, 0, 4, 0],
            [4, 0, -23, 4, 4],
            [0, 4, 4, -23, 4],
            [0, 0, 4, 4, -15]
        ]
        self.assertCountEqual(matrix.tolist(), want)



import itertools
import numpy as np
from qubo_nn.problems.subgraph_isomorphism import SubGraphIsomorphism
from qubo_nn.problems.util import gen_graph


class GraphIsomorphism(SubGraphIsomorphism):
    def __init__(self, cfg, graph1, graph2):
        super(GraphIsomorphism, self).__init__(cfg, graph1, graph2, a=1, b=2)

    @classmethod
    def gen_problems(self, cfg, n_problems, size, seed=None, **kwargs):
        graphs1 = gen_graph(n_problems, size, seed)
        graphs2 = gen_graph(n_problems, size, seed)
        return [
            {"graph1": graph1, "graph2": graph2}
            for graph1, graph2 in zip(graphs1, graphs2)
        ]



import os
import glob
import pickle

import numpy as np
from tensorflow.core.util import event_pb2
from tensorflow.python.lib.io import tf_record


def my_summary_iterator(path):
    for r in tf_record.tf_record_iterator(path):
        yield event_pb2.Event.FromString(r)


def smooth(y, box_pts):
    box = np.ones(box_pts) / box_pts
    y_smooth = np.convolve(y, box, mode='same')
    return y_smooth


def aggregate_single(paths, desired_tag, min_len):
    aggregated = []
    for path in paths:
        path = glob.glob(os.path.join(path, "events.out.tfevents.*"))
        if not path:
            continue
        path = path[0]

        data = []
        for event in my_summary_iterator(path):
            if not event.summary.value:
                continue
            tag = event.summary.value[0].tag
            if tag.endswith(desired_tag):
                val = event.summary.value[0].simple_value
                data.append(val)

        # data = smooth(data, 20)[10:-9]
        # data = smooth(data, 60)[30:-29]
        print(len(data))
        aggregated.append(data)

    if not aggregated:
        return []
    max_len = max(max(len(x) for x in aggregated), min_len)
    aggregated_ = []
    for i, x in enumerate(aggregated):
        aggregated_.append(
            np.pad(
                x,
                (0, max_len - len(x)),
                mode='constant',
                constant_values=(0, x[-1])
            )
        )
    arr = np.array(aggregated_)
    return arr


def aggregate(paths, min_len):
    arr_eval = aggregate_single(paths, 'Loss/Eval', min_len)
    arr_train = aggregate_single(paths, 'Loss/Train', min_len)
    arr_r2 = aggregate_single(paths, 'Custom/R2', min_len)
    return arr_eval, arr_train, arr_r2


def run():
    min_len = 1000
    base_path = '../runs/'
    # problems = ['a19', 'a19_gen_edges', 'mvc3', 'gc1', 'np19_LONG']
    problems = ['a19_2_r2', 'mvc3_r2', 'gc1_r2', 'np19_LONG_r2', 'tsp2_r2']
    kv = {}
    for problem in problems:
        print(problem)
        paths = glob.glob('../runs/*-' + problem)
        data = aggregate(paths, min_len)
        kv[problem] = data

    with open('reverse_losses.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
# n = 6
# color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
# mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)


def gen_table(kv):
    def calc_ci(key, arr):
        arr = arr[~np.isnan(arr)]
        arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        range_ = round(mean - ci[0], 4)
        mean = round(mean, 4)
        return mean, range_

    for k, v in kv.items():
        # arr = arr[:, 1:201]
        v = np.array(v)
        mean, range_ = calc_ci(k, v[0].max(axis=1))  # r2
        print(k, "R2", "%.3f" % mean, "+-", "%.4f" % range_)


def plot(kv):
    tags = {
        'qa_N_64_norm3_5p': '5',
        'qa_N_64_norm3_10p': '10',
        'qa_N_64_norm3_15p': '15',
        'qa_N_64_norm3_20p': '20',
        'qa_N_64_norm3_25p': '25',
        'qa_N_64_norm3_30p': '30',
        'qa_N_64_norm3_35p': '35',
        'qa_N_64_norm3_40p': '40',
        'qa_N_64_norm3_45p': '45',
        'qa_N_64_norm3_50p': '50',
        'qa_N_64_norm3_55p': '55',
        'qa_N_64_norm3_60p': '60',
        'qa_N_64_norm3_65p': '65',
        'qa_N_64_norm3_70p': '70',
        'qa_N_64_norm3_75p': '75',
        'qa_N_64_norm3_80p': '80',
        'qa_N_64_norm3_85p': '85',
        'qa_N_64_norm3_90p': '90'
    }

    def calc_ci(key, arr):
        # arr = arr[~np.isnan(arr)]
        # arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        return mean, mean - ci[0]

    data = []
    err = []
    for k in tags:
        v = kv[k]
        v = np.array(v)
        mean, ci = calc_ci(k, v[0].max(axis=1))
        data.append(mean)
        err.append(ci)

    fig, axs = plt.subplots(1, 1, figsize=(6, 3.5))
    # axs.errorbar(np.arange(len(data)), data, yerr=err, fmt='-o')
    axs.errorbar(np.arange(len(data)), data, yerr=err, fmt='-')
    axs.set_ylabel(r'$R^2$')
    axs.set_xlim(-.2, 17.2)
    axs.set_ylim(0, 1.05)
    axs.set_xticks(np.arange(len(tags)))
    axs.set_xticklabels(list(tags.values()))
    axs.set_xlabel("Percentile removed")

    plt.legend(frameon=False)
    plt.tight_layout()
    plt.show()
    fig.savefig('qa_zeros.png')
    fig.savefig('qa_zeros.pdf')


def run():
    with open('qa_zeros.pickle', 'rb') as f:
        kv = pickle.load(f)
    gen_table(kv)
    plot(kv)


if __name__ == '__main__':
    run()



import os
import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt

from qubo_nn.plots.lib import cmap_mod


NAME = os.path.splitext(os.path.basename(__file__))[0][5:]

mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
# n = 6
# color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
# mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)

PLOT_NAMES = [
    "np", "mc", "mvc", "sp", "m2sat", "spp", "gc", "qa", "qk", "m3sat", "tsp",
    "gi", "sgi", "mcq"
]


def gen_table(kv):
    def calc_ci(key, arr):
        arr = arr[~np.isnan(arr)]
        arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        range_ = round(mean - ci[0], 4)
        mean = round(mean, 4)
        return mean, range_

    for k, v in kv.items():
        if not v:
            continue
        if len(v[0]) == 0:
            continue
        v = np.array(v)
        mean, range_ = calc_ci(k, v[0].max(axis=1))  # r2
        print(k, "R2", "%.3f" % mean, "+-", "%.3f" % range_)


def plot(kv, name):
    fig, axs = plt.subplots(1, 1, figsize=(6, 6))

    tags = {}
    for i in range(1, 20):
        tags["red_" + name + "_" + str(i)] = str(i * 5) + "%"

    def calc_ci(ax, key, arr):
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    for k in tags:
        v = kv[k]
        v = np.array(v)
        calc_ci(axs, k, v[0][:, :400])  # r2

        axs.legend()
        axs.set_ylabel(r'$R^2$')
        axs.set_xlabel("Epoch")

    # plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.legend(frameon=False)
    plt.tight_layout()
    plt.show()
    fig.savefig(NAME + '_' + name + '.png')
    fig.savefig(NAME + '_' + name + '.pdf')


def _get_data(kv, name):
    ret = []

    tags = {}
    for i in range(1, 20):
        tags["red_" + name + "_" + str(i)] = str(i * 5) + "%"

    def calc_ci(key, arr):
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )
        return mean, ci

    for k in tags:
        v = kv[k]
        v = np.array(v)
        x, ci = calc_ci(k, v[0][:, :400])  # r2
        ret.append(x[-1])

    return ret


def plot_all(kv):
    fig, ax = plt.subplots(1, 1, figsize=(9, 6))
    for plot_name in PLOT_NAMES:
        data = _get_data(kv, plot_name)[::-1]
        x = np.arange(len(data))
        ax.plot(x, data, label=plot_name.upper())

    ax.legend()
    ax.set_ylabel(r'$R^2$')
    ax.set_xlabel("Epoch")

    tags = []
    for i in range(1, 20):
        tags.append(str(i * 5) + "%")
    ticks = list(tags[::-1])
    ax.set_xticklabels(ticks)
    ax.set_xticks(np.arange(len(ticks)))

    # plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.legend(frameon=False)
    plt.tight_layout()
    plt.show()
    fig.savefig(NAME + '_all.png')
    fig.savefig(NAME + '_all.pdf')


def plot_matrix(kv):
    fig, ax = plt.subplots(1, 1, figsize=(9, 5))
    all_data = []
    for plot_name in PLOT_NAMES:
        data = _get_data(kv, plot_name)[::-1]
        all_data.append(data)

    tags = []
    for i in range(1, 20):
        tags.append(str(i * 5) + "%")
    ticks = list(tags[::-1])
    ax.set_xticklabels(ticks)
    ax.set_xticks(np.arange(len(ticks)))

    ax.set_yticklabels([p.upper() for p in PLOT_NAMES])
    ax.set_yticks(np.arange(len(PLOT_NAMES)))

    ax.set_xlabel('Hidden layer size as percent of input size')

    im = ax.imshow(all_data, vmin=0, vmax=1, cmap=cmap_mod)
    cbar = ax.figure.colorbar(im, ax=[ax], aspect=30)
    cbar.ax.set_ylabel(r'$R^2$', rotation=-90, va="bottom")
    for i in range(len(PLOT_NAMES)):
        for j in range(len(all_data[0])):
            txt = "%.2f" % round(all_data[i][j], 2)
            col = '#000000' if all_data[i][j] > .65 else '#ffffff'
            ax.text(j, i, txt, ha="center", va="center", color=col, fontsize=7.5)  # noqa

    plt.show()
    fig.savefig(NAME + '_all_matrix.png', bbox_inches='tight')
    fig.savefig(NAME + '_all_matrix.pdf', bbox_inches='tight')


def run():
    with open(NAME + '.pickle', 'rb') as f:
        kv = pickle.load(f)
    # gen_table(kv)
    # for plot_name in PLOT_NAMES:
    #     plot(kv, plot_name)
    #     break

    # plot_all(kv)
    plot_matrix(kv)


if __name__ == '__main__':
    run()



import os
import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt
# from labellines import labelLines

from qubo_nn.plots.lib import cmap_mod


NAME = os.path.splitext(os.path.basename(__file__))[0][5:]

mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
# n = 6
# color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
# mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)

PLOT_TAGS = [
    {
        'sim_pair2_0_0': '0%/0%',
        'sim_pair2_5_5': '5%/5%',
        'sim_pair2_10_10': '10%/10%',
        'sim_pair2_20_20': '20%/20%',
        'sim_pair2_30_30': '30%/30%',
        'sim_pair2_40_30': '40%/30%',
        'sim_pair2_40_40': '40%/40%',
        'sim_pair2_50_40': '50%/40%',
        'sim_pair2_50_50': '50%/50%'
    }
]
PLOT_ACC = [
    {
        'sim_pair2_50_50': (0.3725888888888889, 0.29875555555555555, 0.30924444444444443, 0.35988888888888887, 0.1827888888888889, 0.29852222222222224, 0.24285555555555555, 0.0985111111111111, 0.28524444444444447, 0.29762222222222223, 0.1097111111111111),
        'sim_pair2_50_40': (0.2637111111111111, 0.15051111111111112, 0.12458888888888889, 0.36123333333333335, 0.0021777777777777776, 0.42156666666666665, 0.21152222222222222, 0.20206666666666667, 0.29941111111111113, 0.3038, 0.15123333333333333),
        'sim_pair2_40_40': (0.0, 0.4337, 0.3524111111111111, 0.29794444444444446, 0.24347777777777777, 0.21956666666666666, 0.18303333333333333, 0.259, 0.3559222222222222, 0.3523777777777778, 0.40805555555555556),
        'sim_pair2_40_30': (0.2578666666666667, 0.27166666666666667, 0.2591888888888889, 0.06537777777777778, 0.13322222222222221, 0.31201111111111113, 0.18768888888888888, 0.27355555555555555, 0.3674, 0.33923333333333333, 0.35544444444444445),
        'sim_pair2_30_30': (0.1834111111111111, 0.3353111111111111, 0.1581, 0.0007, 0.11857777777777778, 0.31201111111111113, 0.24482222222222222, 0.2911666666666667, 0.38406666666666667, 0.30445555555555553, 0.3511222222222222),
        'sim_pair2_20_20': (0.33226666666666665, 0.1723222222222222, 0.20721111111111112, 0.1463888888888889, 0.0, 0.0, 0.2715888888888889, 0.3047888888888889, 0.2054111111111111, 0.11474444444444444, 0.24335555555555555),
        'sim_pair2_10_10': (0.11877777777777777, 0.16175555555555557, 0.35057777777777777, 0.06362222222222222, 0.4169, 0.0, 0.0022555555555555554, 0.2995, 0.3517888888888889, 0.12215555555555556, 0.32784444444444444),
        'sim_pair2_5_5': (0.0377, 0.16923333333333335, 0.10471111111111112, 0.24316666666666667, 0.0003111111111111111, 0.29334444444444446, 0.1174, 0.14702222222222222, 0.0, 0.3558, 0.2726),
        'sim_pair2_0_0': (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,)
    }
]
PLOT_NAMES = ['sim2']
PLOT_LIMS = [(0.0, 1.0, 10000)]


def gen_table(kv):
    def calc_ci(key, arr):
        arr = arr[~np.isnan(arr)]
        arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        range_ = round(mean - ci[0], 4)
        mean = round(mean, 4)
        return mean, range_

    for k, v in kv.items():
        if not v:
            continue
        if len(v[0]) == 0:
            continue
        v = np.array(v)
        mean, range_ = calc_ci(k, v[0].min(axis=1))  # r2
        print(k, "R2", "%.3f" % mean, "+-", "%.3f" % range_)


def _get_data(kv, tags, lims):
    def calc_ci(key, arr):
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )
        x = np.arange(len(mean))
        return x, mean, mean - ci[0]

    timeseries_acc_ = [[] for _ in range(11)]
    timeseries_classif = []
    timeseries_classif_err = []
    for k in tags:
        for j, acc in enumerate(PLOT_ACC[0][k]):
            timeseries_acc_[j].append(1 - acc)
        v = kv[k]
        v = np.array(v)
        x, mean, err = calc_ci(k, v[0][:, :lims[-1]])  # r2
        timeseries_classif.append(mean[-1])
        timeseries_classif_err.append(err[-1])

    timeseries_acc_ = np.array(timeseries_acc_)
    _, timeseries_acc, timeseries_acc_err = calc_ci('', timeseries_acc_)

    return timeseries_acc_, timeseries_acc, timeseries_acc_err, timeseries_classif, timeseries_classif_err  # noqa


def plot(kv, tags, name, lims):
    fig, axs = plt.subplots(1, 1, figsize=(8, 4))

    _, timeseries_acc, timeseries_acc_err, timeseries_classif, timeseries_classif_err = _get_data(kv, tags, lims)  # noqa

    x = np.arange(len(timeseries_classif))
    axs.errorbar(x, timeseries_classif, timeseries_classif_err, color=plt.cm.Set2.colors[2])  # noqa
    axs.set_xlabel('Noise factor between pair')
    ticks = list(tags.values())
    ticks.insert(0, '')
    axs.set_xticklabels(ticks)
    axs.set_ylabel('Misclassification rate', color=plt.cm.Set2.colors[2], fontsize=14, fontweight='bold')  # noqa

    axs2 = axs.twinx()
    # print(timeseries_acc_err[0])
    axs2.errorbar(x, timeseries_acc, timeseries_acc_err, color=plt.cm.Set2.colors[4])  # noqa
    # for i, timeseries in enumerate(timeseries_acc):
        # axs2.plot(timeseries, color=plt.cm.Set2.colors[4], label=prob[i])
    axs2.set_ylabel('QUBO Solution Quality', color=plt.cm.Set2.colors[4], fontsize=14, fontweight='bold')  # noqa
    # labelLines(plt.gca().get_lines(), zorder=2.5)

    # plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.xticks(rotation=90)
    # plt.legend(frameon=False)
    # plt.ylim(lims[0:2])
    plt.tight_layout()
    plt.show()
    fig.savefig(NAME + '_' + name + '.png')
    fig.savefig(NAME + '_' + name + '.pdf')


def plot_matrix(kv, tags, lims):
    fig, ax = plt.subplots(1, 1, figsize=(9, 8.5))

    all_data, _, _, _, _ = _get_data(kv, tags, lims)  # noqa

    ticks = list(tags.values())
    ax.set_xticklabels(ticks, rotation=90)
    ax.set_xticks(np.arange(len(ticks)))

    prob = ["NP", "MC", "MVC", "SP", "M2SAT", "SPP", "QA", "QK", "M3SAT",
            "TSP", "MCQ"]
    ax.set_yticklabels(prob)
    ax.set_yticks(np.arange(len(prob)))

    ax.set_xlabel('Noise factor between pair')

    im = ax.imshow(all_data, vmin=0, vmax=1, cmap=cmap_mod)
    for i in range(len(prob)):
        for j in range(len(all_data[0])):
            txt = "%.2f" % round(all_data[i][j], 3)
            ax.text(j, i, txt, ha="center", va="center", color="#000000")
    cbar = ax.figure.colorbar(im, ax=[ax], aspect=30)
    cbar.ax.set_ylabel('Solution quality', rotation=-90, va="bottom")
    for i in range(5):
        plt.axhline(y=i * 2 + 1.5, c='#000000', lw=.5)
    plt.show()
    fig.savefig(NAME + '_all_matrix.png')
    fig.savefig(NAME + '_all_matrix.pdf')


def run():
    with open(NAME + '.pickle', 'rb') as f:
        kv = pickle.load(f)
    gen_table(kv)
    for plot_tags, plot_name, lims in zip(PLOT_TAGS, PLOT_NAMES, PLOT_LIMS):
        plot(kv, plot_tags, plot_name, lims)

    # TODO NOTE: PROBLEM ORDER IS DIFFERENT!!
    # TODO NOTE: PROBLEM ORDER IS DIFFERENT!!
    # TODO NOTE: PROBLEM ORDER IS DIFFERENT!!
    # plot_matrix(kv, PLOT_TAGS[0], PLOT_LIMS[0])


if __name__ == '__main__':
    run()



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len):
    arr_eval = aggregate_single(paths, 'Loss/Eval', min_len)
    arr_train = aggregate_single(paths, 'Loss/Train', min_len)
    arr_r2 = aggregate_single(paths, 'Custom/R2', min_len)
    return arr_eval, arr_train, arr_r2


def run():
    min_len = 500
    base_path = '../runs/'
    problems = ['a19_2_r2', 'a19_2_r2_gen_edges2']
    kv = {}
    for problem in problems:
        print(problem)
        paths = glob.glob('../runs/*-' + problem)
        data = aggregate(paths, min_len)
        kv[problem] = data

    with open('mc_comp_edges.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import random
import itertools
import numpy as np
from qubo_nn.problems.problem import Problem


class SetPacking(Problem):
    def __init__(self, cfg, subset_matrix, P=6):
        self.subset_matrix = subset_matrix
        self.P = P

    def gen_qubo_matrix(self):
        n = self.subset_matrix.shape[1]
        Q = np.zeros((n, n))

        for i in range(n):
            Q[i][i] += 1

        for row in self.subset_matrix:
            idx = list(zip(*np.where(row > 0)))
            tuples = itertools.combinations(idx, 2)
            for j, k in tuples:
                Q[j][k] -= self.P / 2.
                Q[k][j] -= self.P / 2.

            if len(idx) == 1:
                Q[idx[0]][idx[0]] -= self.P / 2.

        return Q

    @classmethod
    def gen_matrix(cls, set_, subsets):
        B = np.zeros((len(set_), len(subsets)))

        for m, x in enumerate(set_):
            for i, subset in enumerate(subsets):
                if x in subset:
                    B[m][i] = 1

        return B

    @classmethod
    def gen_problems(cls, cfg, n_problems, size=(20, 25), **kwargs):
        sorting = cfg["problems"]["SP"].get("sorting", False)

        problems = []

        uniques = set()

        set_ = list(range(size[0]))
        for _ in range(n_problems * 3):
            subsets = set()
            for _ in range(size[1]):
                x = list(filter(lambda x: random.random() < 0.5, set_))
                if not x:
                    continue
                subsets.add(tuple(x))
            if len(subsets) != size[1]:
                continue
            subsets = sorted(list(subsets))
            if tuple(subsets) in uniques:
                continue
            uniques.add(tuple(subsets))

            B = SetPacking.gen_matrix(set_, subsets)

            # Sort it.
            if sorting:
                y = np.array([2 ** i for i in range(len(subsets))])
                z = B @ y
                idx = np.argsort(z)
                B = B[idx]

            problems.append(B)
            if len(problems) == n_problems:
                break

        print("SP generated problems:", len(problems))
        return [{"subset_matrix": matrix} for matrix in problems]



import unittest
import networkx
from qubo_nn.problems import MaxClique


class TestMaxClique(unittest.TestCase):
    def test_gen_qubo_matrix(self):
        """Test whether a correct QUBO is generated.

        Test case from: My brain.
        """
        graph = networkx.Graph(
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)]
        )
        problem = MaxClique({}, graph)
        matrix = problem.gen_qubo_matrix()
        want = [
            [-1.,  0.,  2.,  2.,  2.,  2.],
            [ 0., -1.,  0.,  2.,  2.,  2.],
            [ 2.,  0., -1.,  0.,  2.,  2.],
            [ 2.,  2.,  0., -1.,  0.,  2.],
            [ 2.,  2.,  2.,  0., -1.,  0.],
            [ 2.,  2.,  2.,  2.,  0., -1.]
        ]
        self.assertCountEqual(matrix.tolist(), want)



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, len_, req_len=None):
    fpfn_tot_ratio = aggregate_single(paths, 'Custom/FPFN_TOT_Ratio', min_len, req_len=req_len)
    r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=req_len)
    eval_loss = aggregate_single(paths, 'Loss/Eval', min_len, req_len=req_len)
    return fpfn_tot_ratio, r2, eval_loss


def run():
    min_len = 200
    req_len = 22
    base_paths = ['../runs/', '../runs5/']
    problems = [
        ('qa_N_9_norm2_1M', 10),
        ('qa_N_16_norm1', 10),
        ('qa_N_16_norm2', 10),
        ('qa_N_16_norm3', 10),
        ('qa_N_64_norm1', 10),
        ('qa_N_64_norm2', 10),
        ('qa_N_64_norm3', 10),
        ('qa_N_100_norm1', 10),
        ('qa_N_100_norm2', 10),
        ('qa_N_100_norm3', 10),
        ('qa_N_144_norm3', 10),

        ('qa_N_16_norm2_1M', 10),
        ('qa_N_16_norm2_4M', 10),

        ('qa_special_loss1', 10),

        ('qa_N_9_norm2_1M_2', 10)
    ]
    kv = {}
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        data = aggregate(paths, min_len, problem[1], req_len=req_len)
        kv[problem[0]] = data

    with open('qa_comp.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


np.seterr(divide='ignore', invalid='ignore')

mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')

# Source: https://stackoverflow.com/questions/30079590/
# For continuous:
# n = 10
# color = plt.cm.viridis(np.linspace(0, 1, n))
# mpl.rcParams['axes.prop_cycle'] = cycler.cycler('color', color)
# For discrete:
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)


def plot_loss(kv):
    fig, ax = plt.subplots(figsize=(4, 3))

    tags = {
        'sp4_100k': "6 set size",
        'sp6_100k': "12 set size",
        'sp5_100k': "20 set size"
    }

    def sub_plot_r2(ax, key, arr):
        arr = arr[:, :200]

        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)
        print("Last Loss", key, mean[-1], "+-", mean[-1] - ci[0][-1])

    for k in tags:
        v = kv[k]
        sub_plot_r2(ax, k, v[1])  # This is train loss.
        ax.legend(frameon=False)
        ax.set_ylabel(r'Train Loss')
        ax.set_xlabel("Epoch")
        ax.set_ylim([0, 30])
        # ax.set_title(tags[k])

    plt.tight_layout()
    plt.show()
    fig.savefig('sp_loss.png')
    fig.savefig('sp_loss.pdf')


def plot_vars(kv):
    fig, ax = plt.subplots(figsize=(4, 3))

    tags = {
        'sp4_100k': "6 set size",
        'sp6_100k': "12 set size",
        'sp5_100k': "20 set size"
    }

    def sub_plot_r2(ax, key, arr):
        arr = arr[:, :200]

        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)
        print("Last R2", key, mean[-1], "+-", mean[-1] - ci[0][-1])

    for k in tags:
        v = kv[k]
        sub_plot_r2(ax, k, v[2])  # This is R**2
        ax.legend(frameon=False)
        ax.set_ylabel(r'$R^2$')
        ax.set_xlabel("Epoch")
        ax.set_ylim([0, .5])
        # ax.set_title(tags[k])

    plt.tight_layout()
    plt.show()
    fig.savefig('sp_vars.png')
    fig.savefig('sp_vars.pdf')


def plot_sort(kv):
    fig, axs = plt.subplots(1, 3, figsize=(8.5, 2.5))

    tags = {
        'sp4_100k': "6 set size",
        'sp6_100k': "12 set size",
        'sp5_100k': "20 set size"
    }
    tags_sort = {
        'sp4_sort_100k': "6 set size, sorted",
        'sp6_sort_100k': "12 set size, sorted",
        'sp5_sort_100k': "20 set size, sorted"
    }

    def sub_plot_r2(ax, key, arr, tags):
        arr = arr[:, :300]

        mean = np.mean(arr, axis=0)
        x = np.arange(mean.shape[0])
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)
        print("Last R2", key, mean[-1], "+-", mean[-1] - ci[0][-1])

    for i, (k1, k2) in enumerate(zip(tags.keys(), tags_sort.keys())):
        print(k1, k2)
        ax = axs[i]
        sub_plot_r2(ax, k1, kv[k1][2], tags)
        sub_plot_r2(ax, k2, kv[k2][2], tags_sort)
        ax.legend(frameon=False)
        ax.set_ylabel(r'$R^2$')
        ax.set_xlabel("Epoch")
        ax.set_ylim([0, .6])
        # ax.set_title(tags[k])

    plt.tight_layout()
    plt.show()
    fig.savefig('sp_sorted.png')
    fig.savefig('sp_sorted.pdf')


def run():
    with open('sp_vars.pickle', 'rb') as f:
        kv = pickle.load(f)
    plot_loss(kv)
    plot_vars(kv)
    plot_sort(kv)


if __name__ == '__main__':
    run()



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, cutoff=False):
    arr_eval = aggregate_single(paths, 'Loss/Eval', min_len, req_len=10)
    arr_train = aggregate_single(paths, 'Loss/Train', min_len, req_len=10)
    if cutoff:
        arr_r2 = aggregate_single(paths, 'Custom/R2', min_len, cutoff=6, req_len=10)
    else:
        arr_r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=10)
    return arr_eval, arr_train, arr_r2


def run():
    min_len = 100
    base_path = '../runs/'
    problems = [
        'sp4', 'sp5', 'sp6', 'sp4_sort', 'sp5_sort', 'sp6_sort',
        'sp4_100k', 'sp5_100k', 'sp6_100k', 'sp4_sort_100k', 'sp5_sort_100k', 'sp6_sort_100k',
        'sp4_1M'
    ]
    kv = {}
    for problem in problems:
        print(problem)
        paths = glob.glob(base_path + '*-' + problem)

        print(paths)
        data = aggregate(paths, min_len, False)
        kv[problem] = data

    with open('sp_vars.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import numpy as np
from qubo_nn.problems.problem import Problem


class NumberPartitioning(Problem):
    def __init__(self, cfg, numbers):
        self.numbers = numbers

    def gen_qubo_matrix(self):
        n = len(self.numbers)
        c = sum(self.numbers)

        Q = np.zeros((n, n))
        for i in range(n):
            for j in range(n):
                if i == j:
                    Q[i][j] = self.numbers[i] * (self.numbers[i] - c)
                else:
                    Q[i][j] = self.numbers[i] * self.numbers[j]
        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size=20, **kwargs):
        problems = np.random.randint(0, 100, (n_problems, size))
        return [{"numbers": problem} for problem in problems]



import random
import itertools

import numpy as np
from qubo_nn.problems.problem import Problem


class Max3SAT(Problem):
    def __init__(self, cfg, clauses, n_vars):
        self.clauses = clauses
        self.n_vars = n_vars

    def gen_qubo_matrix(self):
        n = self.n_vars + len(self.clauses)
        Q = np.zeros((n, n))

        # Clauses is a list of 3-tuples of 2-tuples.
        # Eg: [((0, True), (2, True), (3, False)), ((1, True), (3, False), (4, True))]
        # So we loop through all tuples and find the type of it.
        # There are 4 types: (True, True), (True, False), etc.
        for i, clause in enumerate(self.clauses):
            clause_idx = self.n_vars + i
            Q[clause_idx][clause_idx] += 2

            # (1 + w1) * (x1 + x2 + x3) = x1 + x2 + x3 + w1x1 + w1x2 + w1x3
            # (1 + w1) * ((1 - x1) + x2 + x3) = -x1 + x2 + x3 + w1 -w1x1 +w1x2 +w1x3
            # (1 + w1) * ((1 - x1) + x2 + (1 - x3)) = -x1 + x2 -x3 + 2w1 -w1x1 +w1x2 -w1x3
            for item in clause:
                item_idx = item[0]
                val = item[1]
                if val:
                    Q[item_idx][item_idx] -= 1
                    Q[clause_idx][item_idx] -= .5
                    Q[item_idx][clause_idx] -= .5
                if not val:
                    Q[item_idx][item_idx] += 1
                    Q[clause_idx][clause_idx] -= 1
                    Q[clause_idx][item_idx] += .5
                    Q[item_idx][clause_idx] += .5

            # -x1x2 -x1x3 -x2x3
            # -(1-x1)x2 -x1x3 -x2x3 = -1 +x1x2
            # -(1-x1)(1-x2) -x1x3 -x2x3 = -1 -2x1x2 +x1 +x2
            for (item1, item2) in itertools.combinations(clause, 2):
                idx1 = item1[0]
                idx2 = item2[0]
                val1 = item1[1]
                val2 = item2[1]
                if val1 and val2:
                    Q[idx1][idx2] += .5
                    Q[idx2][idx1] += .5
                if not val1 and val2:
                    Q[idx2][idx2] += 1.
                    Q[idx1][idx2] -= .5
                    Q[idx2][idx1] -= .5
                if val1 and not val2:
                    Q[idx1][idx1] += 1.
                    Q[idx1][idx2] -= .5
                    Q[idx2][idx1] -= .5
                if not val1 and not val2:
                    Q[idx1][idx2] += 1.
                    Q[idx2][idx1] += 1.
                    Q[idx1][idx1] -= 1.
                    Q[idx2][idx2] -= 1.

        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size=(20, 25), **kwargs):
        check_duplicates = cfg["problems"]["M3SAT"].get("check_duplicates", False)
        if check_duplicates:
            print("M3SAT gen_problems: Checking for duplicates.")

        # size: First is n_vars, second is number of clauses.
        # Want something like this:
        # [((1, True), (2, True), (3, True)), ((1, True), (2, False), (4, False))]
        n_vars = size[0] - 1
        problems = []
        for _ in range(n_problems):
            problem = []

            # Eg: 25 clauses
            for _ in range(size[1]):
                # Each has THREE tuples!! But random vars and random value.
                # (True/False).
                idx1 = int(round(random.random() * n_vars))
                idx2 = int(round(random.random() * n_vars))
                idx3 = int(round(random.random() * n_vars))
                val1 = bool(random.getrandbits(1))
                val2 = bool(random.getrandbits(1))
                val3 = bool(random.getrandbits(1))

                # TODO: With the checks below we sometimes end up with less than the desired # of clauses.

                if check_duplicates:
                    if idx1 == idx2 or idx1 == idx3 or idx2 == idx3:
                        continue

                clause = (
                    (abs(idx1), val1),
                    (abs(idx2), val2),
                    (abs(idx3), val3)
                )
                clause = sorted(clause)

                if check_duplicates:
                    if clause in problem:
                        continue

                    other_clauses = [
                        [clause[0], (clause[1][0], not clause[1][1]), clause[2]],
                        [clause[0], (clause[1][0], not clause[1][1]), (clause[2][0], not clause[2][1])],
                        [clause[0], clause[1], (clause[2][0], not clause[2][1])],
                        [(clause[0], not clause[0][1]), (clause[1][0], not clause[1][1]), clause[2]],
                        [(clause[0], not clause[0][1]), (clause[1][0], not clause[1][1]), (clause[2][0], not clause[2][1])],
                        [(clause[0], not clause[0][1]), clause[1], (clause[2][0], not clause[2][1])],
                    ]
                    do_cont = False
                    for other_clause in other_clauses:
                        if other_clause in problem:
                            do_cont = True
                    if do_cont:
                        continue

                problem.append(clause)
            if len(problem) == size[1]:
                problems.append(sorted(problem))
        len_ = len(problems)

        return [{"clauses": problem, "n_vars": size[0]} for problem in problems]



import pickle

import numpy as np
import matplotlib.pyplot as plt


with open("qa_debug.pickle", "rb") as f:
    data = pickle.load(f)


losses = [d[0] for d in data]
loss90 = np.percentile(losses, 97.5)
loss10 = np.percentile(losses, 2.5)

avg_diff_x = 0
avg_diff_y = 0
x = []  # good ones
y = []  # bad ones
for d in data:
    if d[0] > loss90:
        y.append((d[3][0], d[2][0], d[1][0]))
        avg_diff_y += abs(np.average(d[3][0][:16]) - np.average(d[3][0][16:]))
    elif d[0] < loss10:
        x.append((d[3][0], d[2][0], d[1][0]))
        avg_diff_x += abs(np.average(d[3][0][:16]) - np.average(d[3][0][16:]))

print(avg_diff_x / 1000., avg_diff_y / 1000.)

idx = np.arange(len(y[0][0]))
width = np.min(np.diff(idx)) / 3

print(len(x), len(y))

for i in range(20):
    print(y[i][2])

    fig, axs = plt.subplots(1, 2, figsize=(12, 4))

    ax = axs[0]
    ax.bar(idx - width + .25, x[i][0], width, label="Want")
    ax.bar(idx + .25, x[i][0], width, label="Predicted")
    ax.set_title("Good loss")
    ax.axhline(y=np.average(x[i][0][:16]), xmin=0, xmax=.5)
    ax.axhline(y=np.average(x[i][0][16:]), xmin=.5, xmax=1.)

    ax = axs[1]
    ax.bar(idx - width + .25, y[i][0], width, label="Want")
    ax.bar(idx + .25, y[i][1], width, label="Predicted")
    ax.set_title("Bad loss")
    ax.axhline(y=np.average(y[i][0][:16]), xmin=0, xmax=.5, c='orange')
    ax.axhline(y=np.average(y[i][0][16:]), xmin=.5, xmax=1., c='orange')

    plt.legend()
    plt.ylabel("Value")
    plt.xlabel("QA Parameter (QUBO size 16x16)")
    plt.tight_layout()
    plt.show()



import numpy as np
from qubo_nn.problems.problem import Problem


class TSP(Problem):
    def __init__(self, cfg, dist_matrix, constraint, P=10):
        self.dist_matrix = dist_matrix
        self.constraint = constraint
        self.P = 10

    def gen_qubo_matrix(self):
        n = len(self.dist_matrix)
        Q = np.zeros((n ** 2, n ** 2))

        quadrants_y = list(range(0, n ** 2, n))
        quadrants_x = quadrants_y[1:] + [quadrants_y[0]]

        # The diagonal positive constraints
        for start_x in quadrants_y:
            for start_y in quadrants_y:
                for i in range(n):
                    Q[start_x + i][start_y + i] = 2 * self.constraint

        # The distance matrices
        for (start_x, start_y) in zip(quadrants_x, quadrants_y):
            for i in range(n):
                for j in range(n):
                    if i == j:
                        continue
                    Q[start_x + i][start_y + j] = self.P * self.dist_matrix[j][i]
                Q[start_x + i][start_y + i] = 2 * self.constraint

        # The middle diagonal negative constraints
        for start_x in quadrants_x:
            for i in range(n):
                Q[start_x + i][start_x + i] = -2 * self.constraint
                for j in range(n):
                    if i != j:
                        Q[start_x + i][start_x + j] += 2 * self.constraint

        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size=4, **kwargs):
        # TODO: 200 and 100 (contraint) is hardcoded !!
        problems = []
        for _ in range(n_problems):
            dist_matrix = np.random.random(size=(size, size))
            np.fill_diagonal(dist_matrix, 0)
            constraint = np.random.random() * 200 + 100
            problems.append({"dist_matrix": dist_matrix.tolist(), "constraint": constraint})
        return problems



import os
import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, max_len, req_len=None):
    # r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=req_len, cutoff=max_len)
    # eval_loss = aggregate_single(paths, 'Loss/Eval', min_len, req_len=req_len, cutoff=max_len)
    train_loss = aggregate_single(paths, 'Loss/Train', min_len, req_len=req_len, cutoff=max_len)
    return [], train_loss


def run():
    min_len = 1000
    base_paths = ['../runs/', '../runs9/']
    problems = [
        ('a3_1', 5),
        ('a3_2', 5000),
        ('a3_3', 5000),
        ('a3ng', 1000),
        ('a3ng_1', 1000),
        ('a3ng_2', 1000),
        ('a3ng_3', 1000),
        ('a3ng_v2', 1000),
        ('a3ng_v2_e1', 100)
    ]
    kv = {}
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        data = aggregate(paths, min_len, 10000000, req_len=problem[1])
        kv[problem[0]] = data

    fname = os.path.splitext(__file__)[0][4:] + '.pickle'
    print("Saving in", fname)
    with open(fname, 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import os
import glob
import pickle

from lib import aggregate_single


problems_short = ["np", "mc", "mvc", "sp", "m2sat", "spp", "gc", "qa", "qk", "m3sat", "tsp", "gi", "sgi", "mcq"]


def aggregate(paths, min_len, max_len, req_len=None):
    r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=req_len, cutoff=max_len)
    eval_loss = aggregate_single(paths, 'Loss/Eval', min_len, req_len=req_len, cutoff=max_len)
    return r2, eval_loss


def run():
    min_len = 1000
    base_paths = ['../runs/']
    problems = []
    for short in problems_short:
        for i in range(1, 20):
            problems.append(
                ("red_" + short + "_" + str(i), 20000)
            )
    kv = {}
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        req_len = 150
        if "gc" in problem[0] or "qa" in problem[0] or "mcq" in problem[0]:
            req_len = 10
        data = aggregate(paths, min_len, problem[1], req_len=req_len)
        kv[problem[0]] = data

    fname = os.path.splitext(__file__)[0][4:] + '.pickle'
    print("Saving in", fname)
    with open(fname, 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import random
import unittest
from qubo_nn.problems import Max2SAT


class TestMax2SAT(unittest.TestCase):
    def test_gen_qubo_matrix(self):
        """Test whether a correct QUBO is generated.

        Test case from: https://arxiv.org/pdf/1811.11538.pdf
        """
        problem = Max2SAT(
            {"problems": {"M2SAT": {}}},
            [
                ((0, True), (1, True)),
                ((0, True), (1, False)),
                ((0, False), (1, True)),
                ((0, False), (1, False)),
                ((0, False), (2, True)),
                ((0, False), (2, False)),
                ((1, True), (2, False)),
                ((1, True), (3, True)),
                ((1, False), (2, True)),
                ((1, False), (2, False)),
                ((2, True), (3, True)),
                ((2, False), (3, False))
            ],
            4
        )
        matrix = problem.gen_qubo_matrix()
        want = [
            [1, 0, 0, 0],
            [0, 0, -.5, .5],
            [0, -.5, 0, 1],
            [0, .5, 1, -2]
        ]
        self.assertCountEqual(matrix.tolist(), want)

    def test_gen_problems(self):
        st0 = random.getstate()
        random.seed(1)
        data = Max2SAT.gen_problems(
            {"problems": {"M2SAT": {}}},
            1,
            size=(20, 25)
        )
        random.setstate(st0)
        self.assertCountEqual(
            data[0]["clauses"],
            [
                [(0, False), (14, False)],
                [(1, False), (16, True)],
                [(1, True), (13, True)],
                [(1, True), (17, False)],
                [(2, True), (6, True)],
                [(3, True), (10, True)],
                [(3, True), (16, False)],
                [(4, False), (4, True)],
                [(4, False), (8, True)],
                [(4, False), (8, True)],
                [(4, False), (14, True)],
                [(4, True), (19, False)],
                [(5, False), (9, False)],
                [(5, False), (15, True)],
                [(6, False), (9, False)],
                [(7, False), (9, False)],
                [(7, True), (13, False)],
                [(10, False), (11, False)],
                [(10, False), (18, True)],
                [(10, True), (15, True)],
                [(11, False), (16, True)],
                [(11, True), (17, True)],
                [(12, False), (15, False)],
                [(13, False), (16, False)],
                [(14, False), (18, True)]
            ]
        )



import itertools
import numpy as np
from qubo_nn.problems.problem import Problem
from qubo_nn.problems.util import gen_graph


class SubGraphIsomorphism(Problem):
    """Generalization of Graph Isomorphism."""
    def __init__(self, cfg, graph1, graph2, a=1, b=2):
        self.graph1 = graph1
        self.graph2 = graph2
        self.a = a
        self.b = b

    def gen_qubo_matrix(self):
        n1 = self.graph1.order()
        n2 = self.graph2.order()

        Q = np.zeros((n1 * n2, n1 * n2))

        for i in range(n1):
            for j in range(n2):
                idx = i * n2 + j
                Q[idx][idx] -= 1 * self.a
            for k, m in itertools.combinations(list(range(n2)), 2):
                idx1 = i * n2 + k
                idx2 = i * n2 + m
                Q[idx1][idx2] += 2 * self.a
                Q[idx2][idx1] += 2 * self.a

        for j in range(n2):
            for i in range(n1):
                idx = i * n2 + j
                Q[idx][idx] -= 1
            for k, m in itertools.combinations(list(range(n1)), 2):
                idx1 = k * n2 + j
                idx2 = m * n2 + j
                Q[idx1][idx2] += 2
                Q[idx2][idx1] += 2

        nodes = list(self.graph1.nodes)
        for edge in self.graph1.edges:
            i = nodes.index(edge[0])
            j = nodes.index(edge[1])

            for i_dash in range(n2):
                idx1 = i * n2 + i_dash
                for j_dash in range(n2):
                    idx2 = j * n2 + j_dash
                    if (i_dash, j_dash) in self.graph2.edges:
                        continue
                    Q[idx1][idx2] += 1 * self.b
                    Q[idx2][idx1] += 1 * self.b
        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size1, size2, seed=None, **kwargs):
        # NOTE: size1 should be smaller than size2 or equal to it.
        graphs1 = gen_graph(n_problems, size1, seed)
        graphs2 = gen_graph(n_problems, size2, seed)
        return [
            {"graph1": graph1, "graph2": graph2}
            for graph1, graph2 in zip(graphs1, graphs2)
        ]



import gzip
import pickle
import numpy as np


with gzip.open('../datasets/1.pickle.gz', 'rb') as f:
    data, labels = pickle.load(f)


# Calc percentiles.
singles = []
medians = []
for i in range(9):
    p = np.percentile(data[int(1e5 * i):int(1e5 * (i + 1))], 90, axis=0)
    medians.append(p)
    p = data[int(1e5 * i) + 100]
    singles.append(p)


with open('qubo_map.pickle', 'wb+') as f:
    pickle.dump((medians, singles), f)



import os
import glob

import numpy as np
import seaborn as sns
import matplotlib.colors as colors
from tensorflow.core.util import event_pb2
from tensorflow.python.lib.io import tf_record


def my_summary_iterator(path):
    for r in tf_record.tf_record_iterator(path):
        yield event_pb2.Event.FromString(r)


def smooth(y, box_pts):
    box = np.ones(box_pts) / box_pts
    y_smooth = np.convolve(y, box, mode='same')
    return y_smooth


def aggregate_single(paths, desired_tag, min_len, cutoff=None, req_len=None):
    aggregated = []
    for path in paths:
        path = glob.glob(os.path.join(path, "events.out.tfevents.*"))
        if not path:
            continue
        path = path[0]

        try:
            data = []
            for event in my_summary_iterator(path):
                if not event.summary.value:
                    continue
                tag = event.summary.value[0].tag
                if tag.endswith(desired_tag):
                    val = event.summary.value[0].simple_value
                    data.append(val)
        except:
            continue

        # data = smooth(data, 20)[10:-9]
        # data = smooth(data, 60)[30:-29]
        if len(data) == 0:
            continue
        if req_len is not None:
            if len(data) <= req_len:
                continue
        print(len(data), data[-1])
        if cutoff is not None:
            data = data[:cutoff]
        aggregated.append(data)

    if not aggregated:
        return []
    max_len = max(max(len(x) for x in aggregated), min_len)
    aggregated_ = []
    for i, x in enumerate(aggregated):
        aggregated_.append(
            np.pad(
                x,
                (0, max_len - len(x)),
                mode='constant',
                constant_values=(0, x[-1])
            )
        )
    arr = np.array(aggregated_)
    return arr


def truncate_colormap2(cmapIn, minval=0.0, maxval=1.0, n=100):
    return colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmapIn.name, a=minval, b=maxval),
        cmapIn(np.linspace(minval, maxval, n)))


cmap_mod = sns.color_palette("mako", as_cmap=True)
cmap_mod = truncate_colormap2(cmap_mod, minval=.0, maxval=1.)



import os
import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


NAME = os.path.splitext(os.path.basename(__file__))[0][5:]

mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
# n = 6
# color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
# mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)

PLOT_TAGS = [
    {
        'm2sat_8x8_10_gen1_3': '16x16',
        'm2sat_8x8_10_gen2_3': '16x16, 14x14',
        'm2sat_8x8_10_gen3_3': '16x16, 14x14, 12x12',
        'm2sat_8x8_10_gen4_3': '16x16, 14x14, 12x12, 10x10',
    }
]
PLOT_NAMES = ['m2sat']
PLOT_LIMS = [(0.0, 1.0, 200)]


def gen_table(kv):
    def calc_ci(key, arr):
        arr = arr[~np.isnan(arr)]
        arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        range_ = round(mean - ci[0], 4)
        mean = round(mean, 4)
        return mean, range_

    for k, v in kv.items():
        if not v:
            continue
        if len(v[0]) == 0:
            continue
        v = np.array(v)
        mean, range_ = calc_ci(k, v[0].max(axis=1))  # r2
        print(k, "R2", "%.3f" % mean, "+-", "%.3f" % range_)


def plot(kv, tags, name, lims):
    fig, axs = plt.subplots(1, 1, figsize=(4, 3))

    def calc_ci(ax, key, arr):
        # arr = arr[~np.isnan(arr)]
        # arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        ax.plot(x, mean, label=tags[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    for k in tags:
        v = kv[k]
        v = np.array(v)
        calc_ci(axs, k, v[0][:, :lims[-1]])  # r2

        axs.legend()
        axs.set_ylabel(r'$R^2$')
        axs.set_xlabel("Epoch")

    # plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.legend(frameon=False)
    plt.ylim(lims[0:2])
    plt.tight_layout()
    plt.show()
    fig.savefig(NAME + '_' + name + '.png')
    fig.savefig(NAME + '_' + name + '.pdf')


def run():
    with open(NAME + '.pickle', 'rb') as f:
        kv = pickle.load(f)
    gen_table(kv)
    for plot_tags, plot_name, lims in zip(PLOT_TAGS, PLOT_NAMES, PLOT_LIMS):
        plot(kv, plot_tags, plot_name, lims)


if __name__ == '__main__':
    run()



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, len_):
    fpfn_tot_ratio = aggregate_single(paths, 'Custom/FPFN_TOT_Ratio', min_len, len_)
    r2 = aggregate_single(paths, 'Custom/R2', min_len, len_)
    return fpfn_tot_ratio, r2


def run():
    min_len = 201
    base_paths = ['../runs3/', '../runs/']
    problems = [
        ('m2sat_16x16_5_F_v2', 201),
        ('m2sat_16x16_10_F_v2', 61),
        ('m2sat_16x16_15_F_v2', 51),
        ('m2sat_16x16_20_F_v2', 51),
        ('m2sat_16x16_25_F_v2', 31),
        ('m2sat_16x16_30_F_v2', 31)
    ]
    kv = {}
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        data = aggregate(paths, min_len, problem[1])
        kv[problem[0]] = data

    with open('m2sat_comp.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, cutoff=False):
    arr_eval = aggregate_single(paths, 'Loss/Eval', min_len)
    arr_train = aggregate_single(paths, 'Loss/Train', min_len)
    if cutoff:
        arr_r2 = aggregate_single(paths, 'Custom/R2', min_len, cutoff=6)
    else:
        arr_r2 = aggregate_single(paths, 'Custom/R2', min_len)
    return arr_eval, arr_train, arr_r2


def run():
    min_len = 100
    base_path = '../runs/'
    problems = ['qa_N_16_norm3', 'qa_N_16_norm3_goddamn']
    kv = {}
    for problem in problems:
        print(problem)
        paths = glob.glob(base_path + '*-' + problem)

        print(paths)
        data = aggregate(paths, min_len, False)
        kv[problem] = data

    with open('qa_ones.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, req_len):
    arr_eval = aggregate_single(paths, 'Loss/Eval', min_len)
    arr_train = aggregate_single(paths, 'Loss/Train', min_len)
    arr_r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=req_len)
    return arr_eval, arr_train, arr_r2


def run():
    min_len = 100
    base_paths = ['../runs/', '../runs5/']
    problems = [
        'm3sat_5_F',
        'm3sat_10_F',
        'm3sat_15V_5_F',
        'm3sat_15V_10_F',
        'm3sat_20V_10_F',
        'm3sat_5V_5_F',
        'm3sat_5V_3_F',
        'm3sat_5V_2_F',

        'm3sat_10V_5_A',
        'm3sat_10V_10_A',
        'm3sat_15V_5_A',
        'm3sat_15V_10_A',
        'm3sat_5V_5_A',
        'm3sat_5V_3_A',
        'm3sat_5V_2_A',
        'm3sat_5V_5_A_2',

        'm2sat_16x16_10_F_v2_1M',
        'm2sat_16x16_5_F_v2_1M',
        'm2sat_16x16_15_F_v2_1M',
        'm2sat_16x16_20_F_v2_1M',
        'm2sat_16x16_25_F_v2_1M',
        'm2sat_16x16_30_F_v2_1M',

        'm2sat_8x8_10_F_v2_1M',
        'm2sat_8x8_5_F_v2_1M',
        'm2sat_8x8_15_F_v2_1M',
        'm2sat_8x8_20_F_v2_1M',
        'm2sat_8x8_25_F_v2_1M',
        'm2sat_8x8_30_F_v2_1M'
    ]

    req_len = 10

    kv = {}
    for problem in problems:
        paths = []
        for base_path in base_paths:
            print(problem)
            paths.extend(glob.glob(base_path + '*-' + problem))

        print(paths)
        data = aggregate(paths, min_len, req_len)
        kv[problem] = data

    with open('m23sat.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import itertools
import numpy as np
from qubo_nn.problems.problem import Problem
from qubo_nn.problems.util import gen_graph


class GraphColoring(Problem):
    def __init__(self, cfg, graph, n_colors, P=4):
        self.graph = graph
        self.n_colors = n_colors
        self.P = P

    def gen_qubo_matrix(self):
        n = self.graph.order() * self.n_colors
        nodes = list(self.graph.nodes)

        Q = np.zeros((n, n))

        for i in range(n):
            Q[i][i] -= self.P

        for i, x in enumerate(nodes):
            cols = [i * self.n_colors + c for c in range(self.n_colors)]
            tuples = itertools.combinations(cols, 2)
            for j, k in tuples:
                Q[j][k] += self.P
                Q[k][j] += self.P

        for edge in self.graph.edges:
            idx1 = nodes.index(edge[0])
            idx2 = nodes.index(edge[1])
            for c in range(self.n_colors):
                idx1c = idx1 * self.n_colors + c
                idx2c = idx2 * self.n_colors + c
                Q[idx1c][idx2c] += self.P / 2.
                Q[idx2c][idx1c] += self.P / 2.
        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size, n_colors, seed=None, **kwargs):
        graphs = gen_graph(n_problems, size, seed)
        return [
            {"graph": graph, "n_colors": n_colors}
            for graph in graphs
        ]



import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
n = 8
color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)


TAGS = {
    'm2sat_16x16_5_F_v2_1M': '5 clauses, 1M',
    'm2sat_16x16_10_F_v2_1M': '10 clauses, 1M',
    'm2sat_16x16_15_F_v2_1M': '15 clauses, 1M',
    'm2sat_16x16_20_F_v2_1M': '20 clauses, 1M',
    'm2sat_16x16_25_F_v2_1M': '25 clauses, 1M',
    'm2sat_16x16_30_F_v2_1M': '30 clauses, 1M'
}
TAGS2 = {
    'm2sat_16x16_5_F_v2': '5 clauses',
    'm2sat_16x16_10_F_v2': '10 clauses',
    'm2sat_16x16_15_F_v2': '15 clauses',
    'm2sat_16x16_20_F_v2': '20 clauses',
    'm2sat_16x16_25_F_v2': '25 clauses',
    'm2sat_16x16_30_F_v2': '30 clauses'
}


def gen_table(kv):
    def calc_ci(key, arr):
        arr = arr[~np.isnan(arr)]
        arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        range_ = round(mean - ci[0], 4)
        mean = round(mean, 4)
        return mean, range_

    for k in TAGS.keys():
        v = kv[k]
        mean, range_ = calc_ci(k, v[2].max(axis=1))  # R2
        print(k, "R2", "%.3f" % mean, "+-", "%.3f" % range_)


def plot(kv, kv2):
    def calc_ci(ax, key, arr, tags, col=None):
        # arr = arr[~np.isnan(arr)]
        # arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        if col is not None:
            ax.plot(x, mean, label=tags[key], c=col)
            ax.fill_between(x, ci[0], ci[1], alpha=.2, color=col)
        else:
            ax.plot(x, mean, label=tags[key])
            ax.fill_between(x, ci[0], ci[1], alpha=.2)

    fig, axs = plt.subplots(1, 1, figsize=(5, 3.0))
    color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]

    for i, k in enumerate(TAGS.keys()):
        v = kv[k]
        calc_ci(axs, k, v[2][:, :10], TAGS, color[i+1])  # R2

        axs.legend()
        axs.set_ylabel(r'$R^2$')
        axs.set_xlabel("Epoch")

    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.ylim((.3, 1.))
    plt.tight_layout()
    plt.show()
    fig.savefig('m2sat_comp_16x16_1M.png')
    fig.savefig('m2sat_comp_16x16_1M.pdf')

    fig, axs = plt.subplots(1, 1, figsize=(5, 3.0))
    color = plt.cm.Blues(np.linspace(.3, 1, n))[::-1]

    for i, k in enumerate(TAGS2.keys()):
        v = kv2[k]
        calc_ci(axs, k, v[1][:, :10], TAGS2, color[i+1])  # R2

        axs.legend()
        axs.set_ylabel(r'$R^2$')
        axs.set_xlabel("Epoch")

    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.ylim((.3, 1.))
    plt.tight_layout()
    plt.show()
    fig.savefig('m2sat_comp_16x16_300k.png')
    fig.savefig('m2sat_comp_16x16_300k.pdf')


def run():
    with open('m23sat.pickle', 'rb') as f:
        kv = pickle.load(f)
    gen_table(kv)

    with open('m2sat_comp.pickle', 'rb') as f:
        kv2 = pickle.load(f)
    plot(kv, kv2)


if __name__ == '__main__':
    run()



import pickle

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from qubo_nn.config import Config
from qubo_nn.pipeline import ReverseOptimizer


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)


kv = {
    "gc1_r2": "21-04-01_13:16:25-2518151-indigiolith.cip.ifi.lmu.de-gc1_r2",
    "tsp2_r2": "21-04-01_13:35:04-9586534-danburit.cip.ifi.lmu.de-tsp2_r2",
    "a19_2_r2": "21-04-01_18:57:45-4208204-datolith.cip.ifi.lmu.de-a19_2_r2",
    "mvc3_r2": "21-04-02_01:17:28-4255422-feueropal.cip.ifi.lmu.de-mvc3_r2",
    "m2sat_16x16_5_F_v2": "21-04-02_07:40:21-3067619-feueropal.cip.ifi.lmu.de-m2sat_16x16_5_F_v2",
    "np19_LONG_r2": "21-04-03_03:56:39-5891150-indigiolith.cip.ifi.lmu.de-np19_LONG_r2",
    "qa_N_100_norm3": "21-04-06_21:10:40-7356487-beryll.cip.ifi.lmu.de-qa_N_100_norm3",
    "sp4": "21-04-09_12:45:49-1875734-feuerachat.cip.ifi.lmu.de-sp4"
}
tags = {
    "gc1_r2": "GC",
    "tsp2_r2": "TSP",
    "a19_2_r2": "MC",
    "mvc3_r2": "MVC",
    "m2sat_16x16_5_F_v2": "M2SAT",
    "np19_LONG_r2": "NP",
    "qa_N_100_norm3": "QA",
    "sp4": "SP"
}


def plot_matrix(cfg_id, ax, axs):
    base_path = '../models/'
    model_fname = base_path + kv[cfg_id]
    with open(base_path + cfg_id + '.pickle', 'rb') as f:
        output_size = pickle.load(f)

    cfg = Config('../').get_cfg(cfg_id)
    optimizer = ReverseOptimizer(cfg, None, None, output_size)
    optimizer.load(model_fname, output_size)

    # for name, param in optimizer.net.named_parameters():
    #     print(name, param)

    # fig, ax = plt.subplots(figsize=(8, 8))
    mat = list(optimizer.net.named_parameters())[0][1].detach().numpy()
    size = min(mat.shape[0], 192)
    im = ax.matshow(mat[:size, :size])
    # ax.get_xaxis().set_visible(False)
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_visible(False)
    ax.set_xlabel(tags[cfg_id])
    return im


fig, axs = plt.subplots(2, 4, figsize=(8, 3), constrained_layout=True)
axs = [ax for sub_ax in axs for ax in sub_ax]
for k, ax in zip(kv, axs):
    im = plot_matrix(k, ax, axs)

cbar = ax.figure.colorbar(im, ax=[axs], aspect=50)
cbar.ax.set_ylabel('Weight Value', rotation=-90, va="bottom")

plt.savefig('weights_all.png')
plt.savefig('weights_all.pdf')
plt.show()



import pickle

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from qubo_nn.config import Config
from qubo_nn.pipeline import ReverseOptimizer


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", plt.cm.Set2.colors)


kv = {
    "gc1_r2": "21-04-01_13:16:25-2518151-indigiolith.cip.ifi.lmu.de-gc1_r2",
    "tsp2_r2": "21-04-01_13:35:04-9586534-danburit.cip.ifi.lmu.de-tsp2_r2",
    "a19_2_r2": "21-04-01_18:57:45-4208204-datolith.cip.ifi.lmu.de-a19_2_r2",
    "mvc3_r2": "21-04-02_01:17:28-4255422-feueropal.cip.ifi.lmu.de-mvc3_r2",
    "m2sat_16x16_5_F_v2": "21-04-02_07:40:21-3067619-feueropal.cip.ifi.lmu.de-m2sat_16x16_5_F_v2",
    "np19_LONG_r2": "21-04-03_03:56:39-5891150-indigiolith.cip.ifi.lmu.de-np19_LONG_r2",
    "qa_N_100_norm3": "21-04-06_21:10:40-7356487-beryll.cip.ifi.lmu.de-qa_N_100_norm3",
    "sp4": "21-04-09_12:45:49-1875734-feuerachat.cip.ifi.lmu.de-sp4"
}


def plot_matrix(cfg_id):
    base_path = '../models/'
    model_fname = base_path + kv[cfg_id]
    with open(base_path + cfg_id + '.pickle', 'rb') as f:
        output_size = pickle.load(f)

    cfg = Config('../').get_cfg(cfg_id)
    optimizer = ReverseOptimizer(cfg, None, None, output_size)
    optimizer.load(model_fname, output_size)

    # for name, param in optimizer.net.named_parameters():
    #     print(name, param)

    fig, ax = plt.subplots(figsize=(8, 8))
    mat = list(optimizer.net.named_parameters())[0][1].detach().numpy()
    size = min(mat.shape[0], 192)
    im = ax.matshow(mat[:size, :size])
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    cbar = ax.figure.colorbar(im, ax=[ax], aspect=40)
    cbar.ax.set_ylabel('Avg Number of Stuns', rotation=-90, va="bottom")

    plt.tight_layout()
    plt.savefig('weights_' + cfg_id + '.png')
    plt.savefig('weights_' + cfg_id + '.pdf')
    plt.show()
    # fig, ax = plt.subplots(figsize=(8, 8))
    # ax.matshow(
    #     list(optimizer.net.named_parameters())[2][1].detach().numpy()[:256, :256],
    #     interpolation='none',
    #     filterrad=25000.
    # )
    # plt.show()


def plot_hist(cfg_id, labels=True):
    base_path = '../models/'
    model_fname = base_path + kv[cfg_id]
    with open(base_path + cfg_id + '.pickle', 'rb') as f:
        output_size = pickle.load(f)

    cfg = Config('../').get_cfg(cfg_id)
    optimizer = ReverseOptimizer(cfg, None, None, output_size)
    optimizer.load(model_fname, output_size)

    fig, ax = plt.subplots(figsize=(6, 6))

    # data = list(optimizer.net.named_parameters())[0][1].detach().numpy().flatten()
    # ax.hist(data, 30000)

    bins = 10000

    data = list(optimizer.net.named_parameters())[0][1].detach().numpy().flatten()
    ax.hist(data, bins)

    if cfg_id == 'a19_2_r2':
        axins = ax.inset_axes([0.1, 0.6, 0.3, 0.3])
        axins.hist(data[np.where(data < -.4)], bins)
        axins.set_xlim(-.51, -.45)
        axins.set_ylim(0, 20)
        if not labels:
            axins.get_xaxis().set_visible(False)
            axins.get_yaxis().set_visible(False)
        ax.indicate_inset_zoom(axins)
    elif cfg_id == 'mvc3_r2':
        axins = ax.inset_axes([0.6, 0.6, 0.3, 0.3])
        axins.hist(data[np.where(data > .17)], bins)
        axins.set_xlim(.17, .23)
        axins.set_ylim(0, 20)
        if not labels:
            axins.get_xaxis().set_visible(False)
            axins.get_yaxis().set_visible(False)
        ax.indicate_inset_zoom(axins)

    if labels:
        plt.ylabel("Number of weights")
        plt.xlabel("Weight value")
    else:
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    plt.tight_layout()
    if labels:
        plt.savefig('weights_hist_' + cfg_id + '2.png')
        plt.savefig('weights_hist_' + cfg_id + '2.pdf')
    else:
        plt.savefig('weights_hist_' + cfg_id + '.png')
        plt.savefig('weights_hist_' + cfg_id + '.pdf')
    plt.show()


for k in kv:
    plot_matrix(k)
    # plot_hist(k, False)



import pickle

import numpy as np
import scipy.stats as st
import matplotlib as mpl
from matplotlib import pyplot as plt


mpl.font_manager._rebuild()
plt.rc('font', family='Raleway')
n = 10
color = plt.cm.Greens(np.linspace(.3, 1, n))[::-1]
mpl.rcParams['axes.prop_cycle'] = plt.cycler('color', color)


TAGS = {
    'm2sat_8x8_5_F_v2_1M': '5 clauses, 1M',
    'm2sat_8x8_10_F_v2_1M': '10 clauses, 1M',
    'm2sat_8x8_15_F_v2_1M': '15 clauses, 1M',
    'm2sat_8x8_20_F_v2_1M': '20 clauses, 1M',
    'm2sat_8x8_25_F_v2_1M': '25 clauses, 1M',
    'm2sat_8x8_30_F_v2_1M': '30 clauses, 1M'
}


def gen_table(kv):
    def calc_ci(key, arr):
        arr = arr[~np.isnan(arr)]
        arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        range_ = round(mean - ci[0], 4)
        mean = round(mean, 4)
        return mean, range_

    for k in TAGS.keys():
        v = kv[k]
        mean, range_ = calc_ci(k, v[2].max(axis=1))  # R2
        print(k, "R2", "%.3f" % mean, "+-", "%.3f" % range_)


def plot(kv):
    fig, axs = plt.subplots(1, 1, figsize=(5, 3.0))

    def calc_ci(ax, key, arr):
        # arr = arr[~np.isnan(arr)]
        # arr = arr[arr != 0.]
        mean = np.mean(arr, axis=0)
        ci = st.t.interval(
            0.95,
            len(arr) - 1,
            loc=np.mean(arr, axis=0),
            scale=st.sem(arr, axis=0)
        )

        x = np.arange(len(mean))

        ax.plot(x, mean, label=TAGS[key])
        ax.fill_between(x, ci[0], ci[1], alpha=.2)

    for k in TAGS.keys():
        v = kv[k]
        calc_ci(axs, k, v[2][:, :50])  # R2

        axs.legend()
        axs.set_ylabel(r'$R^2$')
        axs.set_xlabel("Epoch")

        # # Shrink current axis by 20%
        # box = axs.get_position()
        # axs.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        # # Put a legend to the right of the current axis
        # axs.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    plt.tight_layout()
    plt.show()
    fig.savefig('m2sat_comp_8x8_1M.png')
    fig.savefig('m2sat_comp_8x8_1M.pdf')


def run():
    with open('m23sat.pickle', 'rb') as f:
        kv = pickle.load(f)
    gen_table(kv)
    plot(kv)


if __name__ == '__main__':
    run()



import numpy as np
from qubo_nn.problems.problem import Problem


class QuadraticAssignment(Problem):
    def __init__(self, cfg, flow_matrix, distance_matrix, P=200):
        self.flow_matrix = flow_matrix
        self.distance_matrix = distance_matrix
        self.P = P

        if cfg["problems"]["QA"].get("debug", False):
            self.P = 0.

    def gen_qubo_matrix(self):
        n = len(self.flow_matrix)

        Q = np.zeros((n ** 2, n ** 2))
        for i in range(n):
            for j in range(n):
                for k in range(n):
                    for m in range(n):
                        val = self.flow_matrix[i][j] * self.distance_matrix[k][m]
                        if val == 0:
                            Q[i * n + k][j * n + m] = self.P
                        else:
                            Q[i * n + k][j * n + m] = val

        for i in range(n ** 2):
            Q[i][i] = -self.P * 2

        return Q

    @classmethod
    def gen_problems(self, cfg, n_problems, size=3, **kwargs):
        high = cfg["problems"]["QA"].get("high", 50)  # Outdated.
        debug = cfg["problems"]["QA"].get("debug", False)

        problems = []
        for _ in range(n_problems):
            if debug:
                x = np.arange(1, 50)
            else:
                x = np.arange(1, 1000)

            choice = np.random.choice(x, size=(2, size, size), replace=False)
            flow = choice[0]
            dist = choice[1]

            if debug:
                dist[1][0] = 1.
                flow[1][0] = 1.

            # flow = np.random.randint(low=1, high=high, size=(size, size))
            # dist = np.random.randint(low=1, high=high, size=(size, size))

            np.fill_diagonal(flow, 0)
            np.fill_diagonal(dist, 0)
            problems.append((
                np.tril(flow) + np.tril(flow, -1).T,
                np.tril(dist) + np.tril(dist, -1).T,
            ))
        return [
            {"flow_matrix": flow_matrix, "distance_matrix": distance_matrix}
            for (flow_matrix, distance_matrix) in problems
        ]



import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, len_, req_len=None):
    r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=req_len)
    eval_loss = aggregate_single(paths, 'Loss/Eval', min_len, req_len=req_len)
    return r2, eval_loss


def run():
    min_len = 200
    req_len = 200
    base_paths = ['../runs/']
    problems = [
        ('qa_N_64_norm3_5p', 10),
        ('qa_N_64_norm3_10p', 10),
        ('qa_N_64_norm3_15p', 10),
        ('qa_N_64_norm3_20p', 10),
        ('qa_N_64_norm3_25p', 10),
        ('qa_N_64_norm3_30p', 10),
        ('qa_N_64_norm3_35p', 10),
        ('qa_N_64_norm3_40p', 10),
        ('qa_N_64_norm3_45p', 10),
        ('qa_N_64_norm3_50p', 10),
        ('qa_N_64_norm3_55p', 10),
        ('qa_N_64_norm3_60p', 10),
        ('qa_N_64_norm3_65p', 10),
        ('qa_N_64_norm3_70p', 10),
        ('qa_N_64_norm3_75p', 10),
        ('qa_N_64_norm3_80p', 10),
        ('qa_N_64_norm3_85p', 10),
        ('qa_N_64_norm3_90p', 10)
    ]
    kv = {}
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        data = aggregate(paths, min_len, problem[1], req_len=req_len)
        kv[problem[0]] = data

    with open('qa_zeros.pickle', 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



import os
import glob
import pickle

from lib import aggregate_single


def aggregate(paths, min_len, len_, req_len=None):
    r2 = aggregate_single(paths, 'Custom/R2', min_len, req_len=req_len)
    eval_loss = aggregate_single(paths, 'Loss/Eval', min_len, req_len=req_len)
    return r2, eval_loss


def run():
    min_len = 200
    req_len = 1000
    base_paths = ['../runs/', '../runs7/']
    problems = [
        ('a19_2_generalized_gen4_big_arch', 10),
        ('mvc3_generalized_gen4_big_arch', 10),
        ('gc1_generalized_gen4_big_arch', 10)
        # TODO!!
    ]
    kv = {}
    for problem in problems:
        print(problem)
        paths = []
        for base_path in base_paths:
            paths.extend(glob.glob(base_path + '*-' + problem[0]))
        print(len(paths))
        print(paths)
        data = aggregate(paths, min_len, problem[1], req_len=req_len)
        kv[problem[0]] = data

    fname = os.path.splitext(__file__)[0][4:] + '.pickle'
    print("Saving in", fname)
    with open(fname, 'wb+') as f:
        pickle.dump(kv, f)


if __name__ == '__main__':
    run()



from qubo_nn.problems.number_partitioning import NumberPartitioning
from qubo_nn.problems.max_cut import MaxCut
from qubo_nn.problems.minimum_vertex_cover import MinimumVertexCover
from qubo_nn.problems.set_packing import SetPacking
from qubo_nn.problems.max2sat import Max2SAT
from qubo_nn.problems.set_partitioning import SetPartitioning
from qubo_nn.problems.graph_coloring import GraphColoring
from qubo_nn.problems.quadratic_assignment import QuadraticAssignment
from qubo_nn.problems.quadratic_knapsack import QuadraticKnapsack
from qubo_nn.problems.max3sat import Max3SAT
from qubo_nn.problems.tsp import TSP
from qubo_nn.problems.graph_isomorphism import GraphIsomorphism
from qubo_nn.problems.subgraph_isomorphism import SubGraphIsomorphism
from qubo_nn.problems.max_clique import MaxClique


PROBLEM_REGISTRY = {
    "NP": NumberPartitioning,
    "MC": MaxCut,
    "MVC": MinimumVertexCover,
    "SP": SetPacking,
    "M2SAT": Max2SAT,
    "SPP": SetPartitioning,
    "GC": GraphColoring,
    "QA": QuadraticAssignment,
    "QK": QuadraticKnapsack,
    "M3SAT": Max3SAT,
    "TSP": TSP,
    "GI": GraphIsomorphism,
    "SGI": SubGraphIsomorphism,
    "MCQ": MaxClique
}
